"use strict";
var _a;
Object.defineProperty(exports, "__esModule", { value: true });
const utils_1 = require("@hibas123/utils");
const nanoid = require("nanoid/generate");
const types_1 = require("./types");
const ALPHABET = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
const QueryRequest = Symbol("queryRequest");
const Snapshots = Symbol("snapshots");
// TODO: Request Timeout
class Database {
    constructor(url, database, accesskey, authkey, rootkey) {
        this.initLock = new utils_1.Lock();
        this.requests = new Map();
        this[_a] = new Map();
        const u = new URL(url);
        if (authkey)
            u.searchParams.set("authkey", authkey);
        if (rootkey)
            u.searchParams.set("rootkey", rootkey);
        if (accesskey)
            u.searchParams.set("accesskey", accesskey);
        u.searchParams.set("database", database);
        this.initLock.getLock().then(lock => {
            this.socket = new Database.WebSocket(u.href);
            this.socket.onopen = () => lock.release();
            this.socket.onerror = err => { console.error(err); lock.release(); };
            this.socket.onmessage = msg => {
                let message = JSON.parse(msg.data);
                switch (message.ns) {
                    case "message":
                        this.handleResponse(message.data);
                        break;
                    case "snapshot":
                        this.handleSnapshot(message.data);
                        break;
                    // case "event":
                    //    this.handleEvent(message.data)
                    //    break;
                    case "error_msg":
                        // TODO: Some kind of error handling etc.
                        console.error("Server error:", message.data);
                        break;
                }
            };
        });
    }
    handleSnapshot({ id, data }) {
        let handler = this[Snapshots].get(id);
        if (handler)
            handler(data);
    }
    handleResponse({ id, error, data }) {
        let req = this.requests.get(id);
        if (req) {
            if (error)
                req(new Error(data), undefined);
            else
                req(undefined, data);
        }
        else {
            console.error("Invalid response", id);
        }
    }
    async [(_a = Snapshots, QueryRequest)](type, path, data, id) {
        await this.initLock.getLock().then(l => l.release());
        if (!id)
            id = nanoid(ALPHABET, 21);
        return new Promise((yes, no) => {
            this.requests.set(id, (err, data) => {
                err ? no(err) : yes(data);
                this.requests.delete(id);
            });
            this.socket.send(JSON.stringify({ ns: "query", data: { id, type, path, data } }));
        });
    }
    collection(name) {
        return new CollectionRef(this, [name]);
    }
    async collections() {
        return this[QueryRequest]("collections", [])
            .then(res => res.map(elm => new CollectionRef(this, elm)));
    }
    close() {
        this.socket.close();
    }
}
exports.default = Database;
class Query {
    constructor(db, prefix) {
        this.db = db;
        this.prefix = prefix;
        this._where = [];
    }
    _getNew() {
        let q = new Query(this.db, this.prefix);
        q._where = [...this._where];
        q._limit = this._limit;
        q._orderBy = this._orderBy;
        return q;
    }
    _getDocRef(id) {
        return new DocumentRef(this.db, [...this.prefix, id || nanoid(ALPHABET, 32)]);
    }
    where(fieldPath, opStr, value) {
        let q = this._getNew();
        q._where.push({ fieldPath, opStr, value });
        return q;
    }
    orderBy(fieldPath, direction) {
        let q = this._getNew();
        q._orderBy = { fieldPath, direction };
        return q;
    }
    /**
     * WIP: the functionality is not stable yet!
     * @param limit The number of items
     */
    limit(limit) {
        let q = this._getNew();
        q._limit = limit;
        return q;
    }
    _order(data) {
        if (!this._orderBy)
            return data;
        data.sort((a, b) => {
            //TODO implement!
            return 1;
        });
        return data;
    }
    async get() {
        let data = this._order((await this.db[QueryRequest]("get", this.prefix, {
            limit: this._limit,
            where: this._where
        })).map(e => {
            return {
                id: e.id,
                data: () => e.data,
                ref: this._getDocRef(e.id)
            };
        }));
        return {
            docs: data,
            docChanges: () => {
                return data.map(doc => {
                    return {
                        doc,
                        type: "added"
                    };
                });
            },
            size: data.length,
            empty: data.length <= 0
        };
    }
    async onSnapshot(next) {
        const id = nanoid(ALPHABET, 21);
        let stored = new Map();
        const change = (changes) => {
            let docs = Array.from(stored.values());
            docs = this._order(docs); //TODO: a little inefficient...
            if (this._limit) {
                // TODO: May not be the right way?
                // TODO: Should be limited at all?
                // let old = docs.slice(this._limit);
                // old.forEach(o => stored.delete(o.id));
                docs = docs.slice(0, this._limit);
            }
            next(undefined, {
                docs,
                docChanges: () => changes,
                empty: docs.length <= 0,
                size: docs.length
            });
        };
        this.db[Snapshots].set(id, (data) => {
            let changes = data.map(raw => {
                let doc = {
                    id: raw.id,
                    data: () => raw.data,
                    ref: this._getDocRef(raw.id)
                };
                stored.set(raw.id, doc);
                return {
                    doc,
                    type: raw.type
                };
            });
            change(changes);
        });
        await this.db[QueryRequest]("snapshot", this.prefix, {
            limit: this._limit,
            where: this._where
        }, id).then(data => {
            let docs = data.map(e => ({
                doc: {
                    id: e.id,
                    data: () => e.data,
                    ref: this._getDocRef(e.id)
                },
                type: "added"
            }));
            docs.forEach(doc => {
                stored.set(doc.doc.id, doc.doc);
            });
            change(docs);
        }).catch(err => next(err, undefined));
        return async () => {
            await this.db[QueryRequest]("unsubscribe", this.prefix, undefined, id);
            this.db[Snapshots].delete(id);
            next = undefined;
            stored = undefined;
        };
    }
}
class CollectionRef extends Query {
    get name() {
        return this.prefix[this.prefix.length - 1];
    }
    doc(id) {
        return this._getDocRef(id);
    }
    keys() {
        return this.db[QueryRequest]("keys", this.prefix);
    }
    async add(data) {
        let id = await this.db[QueryRequest]("add", this.prefix, data);
        return new DocumentRef(this.db, [...this.prefix, id]);
    }
    delete() {
        return this.db[QueryRequest]("delete-collection", this.prefix);
    }
}
class DocumentRef {
    constructor(db, prefix) {
        this.db = db;
        this.prefix = prefix;
    }
    get id() {
        return this.prefix[this.prefix.length - 1];
    }
    get() {
        return this.db[QueryRequest]("get", this.prefix);
    }
    set(data) {
        return this.db[QueryRequest]("set", this.prefix, data);
    }
    update(update) {
        let u = {};
        for (const field in update) {
            const data = update[field];
            let res;
            if (typeof data === "object") {
                if (data[types_1.TimestampSymbol]) {
                    res = {
                        type: "timestamp"
                    };
                }
                else if (data[types_1.PushSymbol]) {
                    res = {
                        type: "push",
                        value: data.value
                    };
                }
                else if (data[types_1.IncrementSymbol]) {
                    res = {
                        type: "increment",
                        value: data.value
                    };
                }
                else {
                    res = {
                        type: "value",
                        value: data
                    };
                }
            }
            else {
                res = {
                    type: "value",
                    value: data
                };
            }
            u[field] = res;
        }
        return this.db[QueryRequest]("update", this.prefix, u);
    }
    delete() {
        return this.db[QueryRequest]("delete", this.prefix);
    }
    async onSnapshot(next) {
        let id = nanoid(ALPHABET, 21);
        let stored = undefined;
        const send = () => {
            next(undefined, {
                id: this.id,
                data: () => stored,
                ref: this
            });
        };
        this.db[Snapshots].set(id, (data) => {
            stored = data.data;
            send();
        });
        await this.db[QueryRequest]("snapshot", this.prefix, undefined, id).then((data) => {
            stored = data;
            send();
        }).catch(err => {
            next(err, undefined);
        });
        return async () => {
            await this.db[QueryRequest]("unsubscribe", this.prefix, undefined, id);
            this.db[Snapshots].delete(id);
            next = undefined;
            stored = undefined;
        };
    }
    collection(name) {
        return new CollectionRef(this.db, [...this.prefix, name]);
    }
}
//# sourceMappingURL=database.js.map