import { Lock } from "@hibas123/utils";
import nanoid = require("nanoid/generate");
import { IDocumentRef, ICollectionRef, IDocumentSnapshotHandler, IQuerySnapshot, DocumentChangeType, IQuery, FieldPath, WhereFilterOp, IQuerySnapshotHandler, OrderByDirection, IDocumentChange, IDocumentSnapshot, DocumentUpdate, TimestampSymbol, PushSymbol, IncrementSymbol } from "./types";
const ALPHABET = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";

type QueryTypes = "keys" | "get" | "set" | "update" | "delete" | "add" | "unsubscribe" | "collections" | "delete-collection" | "snapshot";

type SubscribeOptions = Partial<{
   existing: boolean;
}>;

interface Change {
   data: any;
   document: string;
   type: DocumentChangeType;
}

type SubscriptionCallback<T> = (change: {
   type: DocumentChangeType,
   doc: {
      ref: IDocumentRef<T>,
      data: T
   }
}) => void;

type DocRes = {
   id: string,
   data: any
}

const QueryRequest = Symbol("queryRequest");
const Snapshots = Symbol("snapshots");

// TODO: Request Timeout
export default class Database {
   static WebSocket: typeof WebSocket;

   private initLock = new Lock();
   private socket: WebSocket;
   private requests = new Map<string, (err: Error, data: any) => void>();

   constructor(url: string, database: string, accesskey?: string, authkey?: string, rootkey?: string) {
      const u = new URL(url);
      if (authkey)
         u.searchParams.set("authkey", authkey);
      if (rootkey)
         u.searchParams.set("rootkey", rootkey);
      if (accesskey)
         u.searchParams.set("accesskey", accesskey);

      u.searchParams.set("database", database);
      this.initLock.getLock().then(lock => {
         this.socket = new Database.WebSocket(u.href);

         this.socket.onopen = () => lock.release();
         this.socket.onerror = err => { console.error(err); lock.release() };
         this.socket.onmessage = msg => {
            let message: { ns: string, data: any } = JSON.parse(msg.data as string);
            switch (message.ns) {
               case "message":
                  this.handleResponse(message.data);
                  break;
               case "snapshot":
                  this.handleSnapshot(message.data);
                  break;
               // case "event":
               //    this.handleEvent(message.data)
               //    break;
               case "error_msg":
                  // TODO: Some kind of error handling etc.
                  console.error("Server error:", message.data);
                  break;
            }
         }
      })
   }


   [Snapshots] = new Map<string, (data: any) => void>();

   private handleSnapshot({ id, data }: { id: string, data: any }) {
      let handler = this[Snapshots].get(id);
      if (handler)
         handler(data);
   }

   private handleResponse({ id, error, data }: { id: string, error: boolean, data: string }) {
      let req = this.requests.get(id);
      if (req) {
         if (error)
            req(new Error(data), undefined);
         else
            req(undefined, data);
      } else {
         console.error("Invalid response", id);
      }
   }

   public async [QueryRequest]<T = any>(type: QueryTypes, path: string[], data?: any, id?: string): Promise<T> {
      await this.initLock.getLock().then(l => l.release());
      if (!id)
         id = nanoid(ALPHABET, 21);

      return new Promise<any>((yes, no) => {
         this.requests.set(id, (err, data) => {
            err ? no(err) : yes(data)
            this.requests.delete(id);
         });
         this.socket.send(JSON.stringify({ ns: "query", data: { id, type, path, data } }))
      })
   }

   public collection<T = any>(name: string): ICollectionRef<T> {
      return new CollectionRef(this, [name]);
   }

   public async collections(): Promise<ICollectionRef[]> {
      return this[QueryRequest]("collections", [])
         .then(res => res.map(elm => new CollectionRef(this, elm)));
   }

   public close() {
      this.socket.close();
   }
}

class Query implements IQuery {
   private _limit: number;
   private _where: {
      fieldPath: FieldPath,
      opStr: WhereFilterOp,
      value: any
   }[] = [];
   private _orderBy: {
      fieldPath: FieldPath,
      direction: OrderByDirection
   }

   private _getNew() {
      let q = new Query(this.db, this.prefix);

      q._where = [...this._where];
      q._limit = this._limit;
      q._orderBy = this._orderBy;

      return q;
   }

   protected _getDocRef(id?: string): IDocumentRef<any> {
      return new DocumentRef(this.db, [...this.prefix, id || nanoid(ALPHABET, 32)]);
   }

   constructor(protected db: Database, protected prefix: string[]) { }

   public where(fieldPath: FieldPath, opStr: WhereFilterOp, value: any): IQuery {
      let q = this._getNew();
      q._where.push({ fieldPath, opStr, value })
      return q;
   }

   public orderBy(fieldPath: FieldPath, direction: OrderByDirection) {
      let q = this._getNew();
      q._orderBy = { fieldPath, direction }
      return q;
   }


   /**
    * WIP: the functionality is not stable yet!
    * @param limit The number of items
    */
   public limit(limit: number): IQuery {
      let q = this._getNew();
      q._limit = limit;
      return q;
   }

   private _order(data: IDocumentSnapshot[]) {
      if (!this._orderBy)
         return data;

      data.sort((a, b) => {
         //TODO implement!
         return 1;
      })
      return data;
   }

   public async get(): Promise<IQuerySnapshot> {
      let data = this._order((await this.db[QueryRequest]<DocRes[]>("get", this.prefix, {
         limit: this._limit,
         where: this._where
      })).map(e => {
         return <IDocumentSnapshot>{
            id: e.id,
            data: () => e.data,
            ref: this._getDocRef(e.id)
         }
      }))

      return {
         docs: data,
         docChanges: () => {
            return data.map(doc => {
               return <IDocumentChange>{
                  doc,
                  type: "added"
               }
            })
         },
         size: data.length,
         empty: data.length <= 0
      }
   }

   async onSnapshot(next: IQuerySnapshotHandler): Promise<() => void> {
      const id = nanoid(ALPHABET, 21);
      let stored = new Map<string, IDocumentSnapshot>();

      const change = (changes: IDocumentChange[]) => {
         let docs = Array.from(stored.values());
         docs = this._order(docs); //TODO: a little inefficient...
         if (this._limit) {
            // TODO: May not be the right way?
            // TODO: Should be limited at all?
            // let old = docs.slice(this._limit);
            // old.forEach(o => stored.delete(o.id));
            docs = docs.slice(0, this._limit);
         }
         next(undefined, {
            docs,
            docChanges: () => changes,
            empty: docs.length <= 0,
            size: docs.length
         })
      }

      this.db[Snapshots].set(id, (data: (DocRes & { type: DocumentChangeType })[]) => {
         let changes = data.map(raw => {
            let doc = <IDocumentSnapshot>{
               id: raw.id,
               data: () => raw.data,
               ref: this._getDocRef(raw.id)
            };
            stored.set(raw.id, doc);
            return <IDocumentChange>{
               doc,
               type: raw.type
            }
         })
         change(changes);
      });

      await this.db[QueryRequest]<DocRes[]>("snapshot", this.prefix, {
         limit: this._limit,
         where: this._where
      }, id).then(data => {
         let docs = data.map(e => <IDocumentChange>{
            doc: {
               id: e.id,
               data: () => e.data,
               ref: this._getDocRef(e.id)
            },
            type: "added"
         });

         docs.forEach(doc => {
            stored.set(doc.doc.id, doc.doc);
         })
         change(docs);
      }).catch(err => next(err, undefined));


      return async () => {
         await this.db[QueryRequest]("unsubscribe", this.prefix, undefined, id);
         this.db[Snapshots].delete(id);
         next = undefined;
         stored = undefined;
      }
   }
}

class CollectionRef extends Query implements ICollectionRef {
   get name() {
      return this.prefix[this.prefix.length - 1];
   }

   doc(id?: string): IDocumentRef {
      return this._getDocRef(id);
   }

   keys(): Promise<string[]> {
      return this.db[QueryRequest]("keys", this.prefix);
   }

   async add<T = any>(data: T): Promise<IDocumentRef<T>> {
      let id = await this.db[QueryRequest]("add", this.prefix, data);
      return new DocumentRef(this.db, [...this.prefix, id]);
   }

   delete() {
      return this.db[QueryRequest]("delete-collection", this.prefix);
   }
}


class DocumentRef implements IDocumentRef {
   constructor(private db: Database, private prefix: string[]) { }

   get id() {
      return this.prefix[this.prefix.length - 1];
   }

   get(): Promise<any> {
      return this.db[QueryRequest]("get", this.prefix);
   }

   set(data: any): Promise<void> {
      return this.db[QueryRequest]("set", this.prefix, data);
   }

   update(update: DocumentUpdate): Promise<void> {

      let u = {};
      for (const field in update) {
         const data = update[field];
         let res: { type: string, value?: any };
         if (typeof data === "object") {
            if (data[TimestampSymbol]) {
               res = {
                  type: "timestamp"
               }
            } else if (data[PushSymbol]) {
               res = {
                  type: "push",
                  value: data.value
               }
            } else if (data[IncrementSymbol]) {
               res = {
                  type: "increment",
                  value: data.value
               }
            } else {
               res = {
                  type: "value",
                  value: data
               };
            }
         } else {
            res = {
               type: "value",
               value: data
            };
         }

         u[field] = res;
      }

      return this.db[QueryRequest]("update", this.prefix, u);
   }

   delete(): Promise<void> {
      return this.db[QueryRequest]("delete", this.prefix);
   }

   async onSnapshot(next: IDocumentSnapshotHandler<any>) {
      let id = nanoid(ALPHABET, 21);
      let stored: any = undefined;
      const send = () => {
         next(undefined, {
            id: this.id,
            data: () => stored,
            ref: this
         })
      }

      this.db[Snapshots].set(id, (data: Change) => {
         stored = data.data;
         send();
      })


      await this.db[QueryRequest]("snapshot", this.prefix, undefined, id).then((data: any) => {
         stored = data;
         send();
      }).catch(err => {
         next(err, undefined);
      })

      return async () => {
         await this.db[QueryRequest]("unsubscribe", this.prefix, undefined, id);
         this.db[Snapshots].delete(id);
         next = undefined;
         stored = undefined;
      }
   }

   collection(name: string): ICollectionRef {
      return new CollectionRef(this.db, [...this.prefix, name]);
   }
}